#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <errno.h>

#define NUM_OF_PLAYERS 4
#define MAX_CARD_NUM 14
#define CARD_TYPES_NUM 4

#define KARO_STRING "karo"
#define PIK_STRING "pik"
#define TREF_STRING "tref"
#define HERC_STRING "herc"

typedef enum {
    KARO,
    PIK,
    TREF,
    HERC
} CardType;

typedef enum {
    LEFT,
    RIGHT
} Direction;

typedef struct {
    int lastCard;
    CardType cardType;
    Direction direction;
} PlayInfo;

typedef struct {
    long mtype;
    int pid;
} SignInMessage;

typedef struct {
    long mtype;
    int ordinalNumber;
    int semid;
    int shmid;
} ResponseMessage;

void signIn(int msqid) {
    SignInMessage signInMessage;
    signInMessage.mtype = 1;
    signInMessage.pid = getpid();

    if ((msgsnd(msqid, &signInMessage, sizeof(signInMessage) - sizeof(long), 0)) == -1) {
        perror("Ne mogu da posaljem poruku za prijavu");
        exit(1);
    }
}

ResponseMessage receiveConfirmation(int msqid) {
    ResponseMessage responseMessage;
    if ((msgrcv(msqid, &responseMessage, sizeof(responseMessage) - sizeof(long), getpid(), 0)) == -1) {
        perror("Ne mogu da prihvatim poruku za potvrdu prijave");
        exit(1);
    }

    return responseMessage;
}

char* getSignRepresentation(CardType cardType) {
    if (cardType == KARO) {
        return KARO_STRING;
    } else if (cardType == PIK) {
        return PIK_STRING;
    } else if (cardType == TREF) {
        return TREF_STRING;
    } else {
        return HERC_STRING;
    }
}

int main() {

    key_t key = ftok("./mau-mau-player.c", 'a');
    int msgflg = 0644 | IPC_CREAT;
    int msqid = msgget(key, msgflg);

    signIn(msqid);
    ResponseMessage responseMessage = receiveConfirmation(msqid);
    printf("Moj redni broj je: %d/%d\n", responseMessage.ordinalNumber + 1, NUM_OF_PLAYERS);
    // printf("semid = %d, shmid = %d\n", responseMessage.semid, responseMessage.shmid);

    PlayInfo *playInfo = (PlayInfo*) shmat(responseMessage.shmid, NULL, 0);
    
    struct sembuf opr_dec_sem, opr_inc_sem;

    opr_dec_sem.sem_num = responseMessage.ordinalNumber;
    opr_dec_sem.sem_op = -1;
    opr_dec_sem.sem_flg = 0;

    opr_inc_sem.sem_num = responseMessage.ordinalNumber == (NUM_OF_PLAYERS - 1) ? 0 : responseMessage.ordinalNumber + 1;
    opr_inc_sem.sem_op = 1;
    opr_inc_sem.sem_flg = 0;

    srand(getpid());
    while (1) {
        semop(responseMessage.semid, &opr_dec_sem, 1);

        printf("Karta na talonu (pre mog izvlacenja) je: %d (%s)\n", (*playInfo).lastCard, getSignRepresentation((*playInfo).cardType));

        CardType cardType = (CardType) (rand() % CARD_TYPES_NUM);
        int cardNumber = (rand() % MAX_CARD_NUM) + 1;

        printf("Karta koju sam izvukao iz spila je: %d (%s)\n", cardNumber, getSignRepresentation(cardType));

        if ((*playInfo).cardType == cardType || (*playInfo).lastCard == cardNumber) {
            (*playInfo).lastCard = cardNumber;
            (*playInfo).cardType = cardType;
        } else {
            printf("Nemam adekvatnu kartu.\n");
        }

        if ((*playInfo).direction == LEFT) {
            if (cardNumber == 8) {
                opr_inc_sem.sem_num = responseMessage.ordinalNumber == (NUM_OF_PLAYERS - 1) ? 0 : responseMessage.ordinalNumber + 1;
                printf("Menjam smer igre (#notOneDirectionFan) ------>\n");
                (*playInfo).direction = RIGHT;
            } else {
                opr_inc_sem.sem_num = responseMessage.ordinalNumber == 0 ? (NUM_OF_PLAYERS - 1) : responseMessage.ordinalNumber - 1;
            }
        } else {
            if (cardNumber == 8) {
                opr_inc_sem.sem_num = responseMessage.ordinalNumber == 0 ? (NUM_OF_PLAYERS - 1) : responseMessage.ordinalNumber - 1;
                printf("Menjam smer igre (#notOneDirectionFan) <------\n");
                (*playInfo).direction = LEFT;
            } else {
                opr_inc_sem.sem_num = responseMessage.ordinalNumber == (NUM_OF_PLAYERS - 1) ? 0 : responseMessage.ordinalNumber + 1;
            }
        }

        sleep(10);

        semop(responseMessage.semid, &opr_inc_sem, 1);
    }

    return 0;
}