#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <errno.h>

#define NUM_OF_PLAYERS 4
#define MAX_CARD_NUM 14
#define CARD_TYPES_NUM 4

#define KARO_STRING "karo"
#define PIK_STRING "pik"
#define TREF_STRING "tref"
#define HERC_STRING "herc"

typedef enum {
    KARO,
    PIK,
    TREF,
    HERC
} CardType;

typedef enum {
    LEFT,
    RIGHT
} Direction;

typedef struct {
    int lastCard;
    CardType cardType;
    Direction direction;
} PlayInfo;

typedef struct {
    long mtype;
    int pid;
} SignInMessage;

typedef struct {
    long mtype;
    int ordinalNumber;
    int semid;
    int shmid;
} ResponseMessage;

union semun {
    int val;
    struct semid_ds *buf;
    unsigned short *array;
};

SignInMessage receiveSignIn(int msqid) {
    SignInMessage signInMessage;

    if ((msgrcv(msqid, &signInMessage, sizeof(signInMessage) - sizeof(long), 1, 0)) == -1) {
        perror("Ne mogu da prihvatim poruku za prijavu poruku");
        exit(1);
    }

    return signInMessage;
}

void sendConfirmationMessage(int msqid, ResponseMessage responseMessage) {
    if ((msgsnd(msqid, &responseMessage, sizeof(responseMessage) - sizeof(long), 0)) == -1) {
        perror("Ne mogu da posaljem poruku za potvrdu prijave");
        exit(1);
    }
}

char* getSignRepresentation(CardType cardType) {
    if (cardType == KARO) {
        return KARO_STRING;
    } else if (cardType == PIK) {
        return PIK_STRING;
    } else if (cardType == TREF) {
        return TREF_STRING;
    } else {
        return HERC_STRING;
    }
}

int main() {

    key_t key = ftok("./mau-mau-player.c", 'a');
    int msgflg = 0644 | IPC_CREAT;
    int msqid = msgget(key, msgflg);

    int currPlayerOrdinal = 0;
    int *players = (int*)malloc(NUM_OF_PLAYERS * sizeof(int));

    while (1) {
        SignInMessage signInMessage = receiveSignIn(msqid);
        players[currPlayerOrdinal++] = signInMessage.pid;
        printf("Igrac %d se prijavio za igru.\n", signInMessage.pid);

        if (currPlayerOrdinal == NUM_OF_PLAYERS) {
            currPlayerOrdinal = 0;
            
            pid_t pid = fork();
            if (pid == -1) {
                perror("Ne mogu da fork-ujem proces");
                exit(1);
            } else if (pid == 0) {
                printf("Zapocenjem novu igru! (proces %d)\n", getpid());
                srand(getpid());

                int semnum = NUM_OF_PLAYERS;
                int semflg = 0644 | IPC_CREAT;
                int semid = semget(IPC_PRIVATE, semnum, semflg);

                union semun sem_init;
                sem_init.array = (unsigned short*)malloc(NUM_OF_PLAYERS * sizeof(unsigned short));
                sem_init.array[0] = 1;
                sem_init.array[1] = 0;
                sem_init.array[2] = 0;
                sem_init.array[3] = 0;

                semctl(semid, 0, SETALL, sem_init);

                int shmflg = 0644 | IPC_CREAT;
                int shmid = shmget(IPC_PRIVATE, sizeof(PlayInfo), shmflg);

                PlayInfo *playInfo = (PlayInfo*) shmat(shmid, NULL, 0);
                (*playInfo).lastCard = (rand() % MAX_CARD_NUM) + 1;
                (*playInfo).cardType = (CardType) (rand() % CARD_TYPES_NUM);
                (*playInfo).direction = RIGHT;

                printf("Pocetna karta na talonu: %d (%s)\n", (*playInfo).lastCard, getSignRepresentation((*playInfo).cardType));

                ResponseMessage responseMessage;
                responseMessage.semid = semid;
                responseMessage.shmid = shmid;

                for (int i = 0; i < NUM_OF_PLAYERS; i++) {
                    responseMessage.mtype = players[i];
                    responseMessage.ordinalNumber = i;

                    sendConfirmationMessage(msqid, responseMessage);
                }

                // for (;;); // uneccessary
            }
        }
    }

    return 0;
}